/* $NetBSD$ */

/*
 * Copyright (c) 2009 Arnaud Ysmal.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/dirent.h>
#include <sys/fstypes.h>
#include <sys/mount.h>
#include <sys/sysctl.h>

#include <dirent.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <mntopts.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <prop/proplib.h>

#include <rump/rump.h>
#include <rump/rump_syscalls.h>

#include "prop_mount.h"
#include "internal.h"

static int specific_getmntopts(prop_dictionary_t, char *);
static ssize_t get_vfstypes(char *, size_t);

static const struct mntopt mopts[] = {
	MOPT_STDOPTS,
	MOPT_FORCE,
	MOPT_UPDATE,
	MOPT_NULL
};

int
prop_mount(int *argc, char **argv[])
{
	struct plistref plref;
	prop_dictionary_t dict;
	mntoptparse_t mp;
	int ch, mntflags, rv;
	char *fs, *outputfile, sfs[1024], *fspec, *progname;

	mntflags = rv = 0;
	dict = NULL;
	fs = outputfile = fspec = NULL;
	progname = (*argv)[0];
	
	if (*argc == 1)
		return -1;
	
	while ((ch = getopt(*argc, *argv, "f:o:s:t:w:")) != -1)
		switch (ch) {
		case 'f': /* load dictionary from a file */
			if (dict != NULL)
				errx(EXIT_FAILURE,
				    "-f should be set before -s");
			dict = prop_dictionary_internalize_from_file(optarg);
			if (dict == NULL) {
				warn("prop_dictionary_create");
				return -1;
			}
			break;
		case 's': /* use specific fs mount options */
			if (dict == NULL) {
				dict = prop_dictionary_create();
				if (dict == NULL) {
					warn("prop_dictionary_create");
					return -1;
				}
			}
			specific_getmntopts(dict, optarg);
			break;
		case 'o': /* use generic mount options */
			mp = getmntopts(optarg, mopts, &mntflags, 0);
			if (mp == NULL)
				return -1;
			freemntopts(mp);
			break;
		case 't': /* use a given file system */
			fs = optarg;
			break;
		case 'w': /* write dictionary to a file */
			outputfile = optarg;
			break;
		default:
			return -1;
		}

	(*argv) += optind - 1;
	(*argc) -= optind - 1;
	(*argv)[0] = progname;

	optind = optreset = 1;

	/*
	 * fspec from file or given as "-o fspec=filename"
	 */
	if (dict != NULL)
		prop_dictionary_get_cstring(dict, "fspec", &fspec);

	if (fspec == NULL && *argc <= 1)
		return -1;

	if (fspec == NULL) {
		fspec = (*argv)[1];
		if (dict == NULL) {
			dict = prop_dictionary_create();
			if (dict == NULL) {
				warn("prop_dictionary_create");
				return -1;
			}
		}
		if (!prop_dictionary_set_cstring(dict, "fspec", (*argv)[1])) {
			warn("prop_dictionary_set_cstring");
			return -1;
		}
		(*argv)[1] = progname;
		++(*argv);
		--(*argc);
	}

	/* using plistref to pass the dictionary to the mount syscall */
	plref.pref_plist = prop_dictionary_externalize(dict);
	plref.pref_len = strlen(plref.pref_plist) + 1;

	/* if -w, write the dictionary to outputfile */
	if (outputfile != NULL)
		prop_dictionary_externalize_to_file(dict, outputfile);

	rump_init();
	fsu_ukfs_modload_dir("/usr/lib");

	rump_fakeblk_register(fspec);
	if (fs == NULL && get_vfstypes(sfs, sizeof(sfs)) != 0) {
		/* the file system was not given, try all supported */
		fs = strtok(sfs, " ");
		do {
			printf("Trying %s: ", fs);
			rv = rump_sys_mount(fs, "/", mntflags,
			    &plref, sizeof(plref));
		} while ((fs = strtok(NULL, " ")) != NULL && rv != 0);
	} else if (fs != NULL)
		rv = rump_sys_mount(fs, "/", mntflags, &plref, sizeof(plref));
	else
		rv = -1;
	
	rump_fakeblk_deregister(fspec);
	prop_object_release(dict);
	return rv;
}

/* XXX: now rump allows to mkdir mount points */
void
prop_unmount(void)
{

	rump_sys_unmount("/", MNT_FORCE);
}

const char *
fsu_mount_usage(void)
{

	return "[-t fstype] [-o generic opts] [-s fs specific opts] "
	    "[-w outputfile.xml] [-f inputfile.xml] [device]";
}

/*
 * proplib uses 64bits int for numbers
 */
int
specific_getmntopts(prop_dictionary_t dict, char *opts)
{
	char *opt, *sep, *ep;
	int64_t tmp;
	bool bval;
	
	opt = strtok(opts, ",");
	do {
		sep = strchr(opt, '=');
		if (sep != NULL && *(sep + 1) == '\0') {
			*sep = '\0';
			sep = NULL;
		}
		if (sep != NULL) {
			*sep++ = 0;
			if (strstr(opt, "uid") != NULL)
				tmp = a_uid(sep);
			else if (strstr(opt, "gid") != NULL)
				tmp = a_gid(sep);
			else if (strstr(opt, "mode") != NULL ||
			    strstr(opt, "mask") != NULL)
				tmp = a_mask(sep);
			else {
				tmp = strtoll(sep, &ep, 0);
				if (*ep != '\0' ||
				    ((tmp == LLONG_MIN || tmp == LLONG_MAX) &&
					errno == ERANGE)) {
					prop_dictionary_set_cstring(dict,
					    opt, sep);
#ifdef DEBUG
					printf("Adding (str): %s = %s\n",
					    opt, sep);
#endif
					continue;
				}
			}
#ifdef DEBUG
			printf("Adding (int): %s = %lld\n", opt, tmp);
#endif
			if (tmp >= 0)
				prop_dictionary_set_uint64(dict, opt, tmp);
			else
				prop_dictionary_set_int64(dict, opt, tmp);

		} else {
			if (*opt == 'n' && *(opt + 1) == 'o') {
				/* XXX: handle options starting with no */
#ifdef DEBUG
				printf("Adding (bool): %s = true\n", opt);
#endif	
				prop_dictionary_set_bool(dict, opt, 1);
				bval = 0;
				opt += 2;
			} else
				bval = 1;
#ifdef DEBUG
			printf("Adding (bool): %s = %s\n", opt,
			    bval == 1 ? "true" : "false");
#endif
			prop_dictionary_set_bool(dict, opt, bval);
		}
	} while ((opt = strtok(NULL, ",")) != NULL);

	return 0;
}

/* function coming from libukfs */
/*
 * Copyright (c) 2007, 2008  Antti Kantee.  All Rights Reserved.
 *
 * Development of this software was supported by the
 * Finnish Cultural Foundation.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
ssize_t
get_vfstypes(char *buf, size_t buflen)
{
	int mib[3];
	struct sysctlnode q, ans[128];
	size_t alen;
	unsigned int i;

	mib[0] = CTL_VFS;
	mib[1] = VFS_GENERIC;
	mib[2] = CTL_QUERY;
	alen = sizeof(ans);

	memset(&q, 0, sizeof(q));
	q.sysctl_flags = SYSCTL_VERSION;

	if (rump_sys___sysctl(mib, 3, ans, &alen, &q, sizeof(q)) == -1) {
		return -1;
	}

	for (i = 0; i < alen/sizeof(ans[0]); i++)
		if (strcmp("fstypes", ans[i].sysctl_name) == 0)
			break;
	if (i == alen/sizeof(ans[0])) {
		errno = ENXIO;
		return -1;
	}

	mib[0] = CTL_VFS;
	mib[1] = VFS_GENERIC;
	mib[2] = ans[i].sysctl_num;

	if (rump_sys___sysctl(mib, 3, buf, &buflen, NULL, 0) == -1) {
		return -1;
	}

	return buflen;
}
